#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j import GraphDatabase
from neo4j.exceptions import (
    Neo4jError,
    TransientError,
)

from tests.stub.conftest import StubCluster

# python -m pytest tests/stub/test_accesslevel.py -s -v


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/return_1_in_read_tx.script"),
        ("v4x0/router.script", "v4x0/tx_return_1_port_9004.script"),
    ]
)
def test_read_transaction(driver_info, test_scripts):
    # python -m pytest tests/stub/test_accesslevel.py -s -v -k test_read_transaction
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:

                def unit_of_work(tx):
                    total = 0
                    for record in tx.run("RETURN 1"):
                        total += record[0]
                    return total

                value = session.read_transaction(unit_of_work)
                assert value == 1


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/return_1_in_write_tx.script"),
        ("v4x0/router.script", "v4x0/tx_return_1_port_9006.script"),
    ]
)
def test_write_transaction(driver_info, test_scripts):
    # python -m pytest tests/stub/test_accesslevel.py -s -v -k test_write_transaction
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:

                def unit_of_work(tx):
                    total = 0
                    for record in tx.run("RETURN 1"):
                        total += record[0]
                    return total

                value = session.write_transaction(unit_of_work)
                assert value == 1


@pytest.mark.skip(reason="BROKEN, This test is broken and should be an integration test.")
@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/error_in_read_tx.script"),
        ("v4x0/router.script", "v4x0/tx_run_with_failure_syntax_error_port_9004.script"),
    ]
)
def test_read_transaction_with_error(driver_info, test_scripts):
    # python -m pytest tests/stub/test_accesslevel.py -s -v -k test_read_transaction_with_error
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:

                def unit_of_work(tx):
                    tx.run("X")

                with pytest.raises(Neo4jError):
                    _ = session.read_transaction(unit_of_work)


@pytest.mark.skip(reason="BROKEN, This test is broken and should be an integration test.")
@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/error_in_write_tx.script"),
        ("v4x0/router.script", "v4x0/tx_run_with_failure_syntax_error_port_9006.script"),
    ]
)
def test_write_transaction_with_error(driver_info, test_scripts):
    # python -m pytest tests/stub/test_accesslevel.py -s -v -k test_write_transaction_with_error
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:

                def unit_of_work(tx):
                    tx.run("X")

                with pytest.raises(Neo4jError):
                    _ = session.write_transaction(unit_of_work)


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/return_1_in_read_tx_twice.script"),
        ("v4x0/router.script", "v4x0/tx_two_subsequent_return_1_port_9004.script"),
    ]
)
def test_two_subsequent_read_transactions(driver_info, test_scripts):
    # python -m pytest tests/stub/test_accesslevel.py -s -v -k test_two_subsequent_read_transactions
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:

                def unit_of_work(tx):
                    total = 0
                    for record in tx.run("RETURN 1"):
                        total += record[0]
                    return total

                value = session.read_transaction(unit_of_work)
                assert value == 1
                value = session.read_transaction(unit_of_work)
                assert value == 1


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/return_1_in_write_tx_twice.script"),
        ("v4x0/router.script", "v4x0/tx_two_subsequent_return_1_port_9006.script"),
    ]
)
def test_two_subsequent_write_transactions(driver_info, test_scripts):
    # python -m pytest tests/stub/test_accesslevel.py -s -v -k test_two_subsequent_write_transactions
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:

                def unit_of_work(tx):
                    total = 0
                    for record in tx.run("RETURN 1"):
                        total += record[0]
                    return total

                value = session.write_transaction(unit_of_work)
                assert value == 1
                value = session.write_transaction(unit_of_work)
                assert value == 1


@pytest.mark.skip(reason="BOOKMARK, AttributeError: 'Session' object has no attribute 'last_bookmark'")
@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/return_1_in_read_tx.script", "v3/return_2_in_write_tx.script"),
        ("v4x0/router.script", "v4x0/tx_return_1_port_9004.script", "v4x0/tx_return_2_with_bookmark_port_9006.script"),
    ]
)
def test_read_tx_then_write_tx(driver_info, test_scripts):
    # python -m pytest tests/stub/test_accesslevel.py -s -v -k test_read_tx_then_write_tx
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:

                def unit_of_work_1(tx):
                    total = 0
                    for record in tx.run("RETURN 1"):
                        total += record[0]
                    return total

                def unit_of_work_2(tx):
                    total = 0
                    for record in tx.run("RETURN 2"):
                        total += record[0]
                    return total

                value = session.read_transaction(unit_of_work_1)
                assert session.last_bookmark() == "bookmark:1"
                assert value == 1
                value = session.write_transaction(unit_of_work_2)
                assert session.last_bookmark() == "bookmark:2"
                assert value == 2


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/return_1_in_write_tx.script", "v3/return_2_in_read_tx.script"),
        ("v4x0/router.script", "v4x0/tx_return_1_port_9006.script", "v4x0/tx_return_2_with_bookmark_port_9004.script"),
    ]
)
def test_write_tx_then_read_tx(driver_info, test_scripts):
    # python -m pytest tests/stub/test_accesslevel.py -s -v -k test_write_tx_then_read_tx
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:

                def unit_of_work_1(tx):
                    total = 0
                    for record in tx.run("RETURN 1"):
                        total += record[0]
                    return total

                def unit_of_work_2(tx):
                    total = 0
                    for record in tx.run("RETURN 2"):
                        total += record[0]
                    return total

                value = session.write_transaction(unit_of_work_1)
                assert value == 1
                value = session.read_transaction(unit_of_work_2)
                assert value == 2


@pytest.mark.skip(reason="BROKEN")
@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/user_canceled_read.script"),
        ("v4x0/router.script", "v4x0/tx_return_1_reset_port_9004.script"),
    ]
)
def test_no_retry_read_on_user_canceled_tx(driver_info, test_scripts):
    # python -m pytest tests/stub/test_accesslevel.py -s -v -k test_no_retry_read_on_user_canceled_tx
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:
                def unit_of_work(tx):
                    tx.run("RETURN 1")

                with pytest.raises(TransientError):
                    _ = session.read_transaction(unit_of_work)


@pytest.mark.skip(reason="BROKEN")
@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/user_canceled_write.script"),
        ("v4x0/router.script", "v4x0/tx_return_1_reset_port_9006.script"),
    ]
)
def test_no_retry_write_on_user_canceled_tx(driver_info, test_scripts):
    # python -m pytest tests/stub/test_accesslevel.py -s -v -k test_no_retry_write_on_user_canceled_tx
    with StubCluster(*test_scripts):
        with GraphDatabase.driver(driver_info["uri_neo4j"], auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:
                def unit_of_work(tx):
                    tx.run("RETURN 1")

                with pytest.raises(TransientError):
                    _ = session.write_transaction(unit_of_work)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from neo4j.graph import (
    Node,
    Relationship,
    Path,
)


def test_node(cypher_eval):
    a = cypher_eval("CREATE (a:Person {name:'Alice'}) "
                    "RETURN a")
    assert isinstance(a, Node)
    assert set(a.labels) == {"Person"}
    assert dict(a) == {"name": "Alice"}


def test_relationship(cypher_eval):
    a, b, r = cypher_eval("CREATE (a)-[r:KNOWS {since:1999}]->(b) "
                          "RETURN [a, b, r]")
    assert isinstance(r, Relationship)
    assert r.type == "KNOWS"
    assert dict(r) == {"since": 1999}
    assert r.start_node == a
    assert r.end_node == b


def test_path(cypher_eval):
    a, b, c, ab, bc, p = cypher_eval("CREATE p=(a)-[ab:X]->(b)-[bc:X]->(c) "
                                     "RETURN [a, b, c, ab, bc, p]")
    assert isinstance(p, Path)
    assert len(p) == 2
    assert p.nodes == (a, b, c)
    assert p.relationships == (ab, bc)
    assert p.start_node == a
    assert p.end_node == c



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from uuid import uuid4

import neo4j.api
from neo4j.data import DataDehydrator
from neo4j.exceptions import (
    ConfigurationError,
)

standard_ascii = [chr(i) for i in range(128)]
not_ascii = "♥O◘♦♥O◘♦"

# python -m pytest tests/unit/test_api.py -s


def dehydrated_value(value):
    return DataDehydrator.fix_parameters({"_": value})["_"]


def test_value_dehydration_should_allow_none():
    # python -m pytest tests/unit/test_api.py -s -k test_value_dehydration_should_allow_none
    assert dehydrated_value(None) is None


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (True, True),
        (False, False),
    ]
)
def test_value_dehydration_should_allow_boolean(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_value_dehydration_should_allow_boolean
    assert dehydrated_value(test_input) is expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (0, 0),
        (1, 1),
        (0x7F, 0x7F),
        (0x7FFF, 0x7FFF),
        (0x7FFFFFFF, 0x7FFFFFFF),
        (0x7FFFFFFFFFFFFFFF, 0x7FFFFFFFFFFFFFFF),
    ]
)
def test_value_dehydration_should_allow_integer(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_value_dehydration_should_allow_integer
    assert dehydrated_value(test_input) == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (0x10000000000000000, ValueError),
        (-0x10000000000000000, ValueError),
    ]
)
def test_value_dehydration_should_disallow_oversized_integer(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_value_dehydration_should_disallow_oversized_integer
    with pytest.raises(expected):
        dehydrated_value(test_input)


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (0.0, 0.0),
        (-0.1, -0.1),
        (3.1415926, 3.1415926),
        (-3.1415926, -3.1415926),
    ]
)
def test_value_dehydration_should_allow_float(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_value_dehydration_should_allow_float
    assert dehydrated_value(test_input) == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (u"", u""),
        (u"hello, world", u"hello, world"),
        ("".join(standard_ascii), "".join(standard_ascii)),
    ]
)
def test_value_dehydration_should_allow_string(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_value_dehydration_should_allow_string
    assert dehydrated_value(test_input) == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (bytearray(), bytearray()),
        (bytearray([1, 2, 3]), bytearray([1, 2, 3])),
    ]
)
def test_value_dehydration_should_allow_bytes(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_value_dehydration_should_allow_bytes
    assert dehydrated_value(test_input) == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        ([], []),
        ([1, 2, 3], [1, 2, 3]),
        ([1, 3.1415926, "string", None], [1, 3.1415926, "string", None])
    ]
)
def test_value_dehydration_should_allow_list(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_value_dehydration_should_allow_list
    assert dehydrated_value(test_input) == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        ({}, {}),
        ({u"one": 1, u"two": 1, u"three": 1}, {u"one": 1, u"two": 1, u"three": 1}),
        ({u"list": [1, 2, 3, [4, 5, 6]], u"dict": {u"a": 1, u"b": 2}}, {u"list": [1, 2, 3, [4, 5, 6]], u"dict": {u"a": 1, u"b": 2}}),
        ({"alpha": [1, 3.1415926, "string", None]}, {"alpha": [1, 3.1415926, "string", None]}),
    ]
)
def test_value_dehydration_should_allow_dict(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_value_dehydration_should_allow_dict
    assert dehydrated_value(test_input) == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (object(), TypeError),
        (uuid4(), TypeError),
    ]
)
def test_value_dehydration_should_disallow_object(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_value_dehydration_should_disallow_object
    with pytest.raises(expected):
        dehydrated_value(test_input)


def test_bookmark_initialization_with_no_values():
    # python -m pytest tests/unit/test_api.py -s -k test_bookmark_initialization_with_no_values
    bookmark = neo4j.api.Bookmark()
    assert bookmark.values == frozenset()
    assert bool(bookmark) is False
    assert repr(bookmark) == "<Bookmark values={}>"


@pytest.mark.parametrize(
    "test_input, expected_values, expected_bool, expected_repr",
    [
        ((None,), frozenset(), False, "<Bookmark values={}>"),
        ((None, None), frozenset(), False, "<Bookmark values={}>"),
        (("bookmark1", None), frozenset({"bookmark1"}), True, "<Bookmark values={'bookmark1'}>"),
        (("bookmark1", None, "bookmark2", None), frozenset({"bookmark1", "bookmark2"}), True, "<Bookmark values={'bookmark1', 'bookmark2'}>"),
        ((None, "bookmark1", None, "bookmark2", None, None, "bookmark3"), frozenset({"bookmark1", "bookmark2", "bookmark3"}), True, "<Bookmark values={'bookmark1', 'bookmark2', 'bookmark3'}>"),
    ]
)
def test_bookmark_initialization_with_values_none(test_input, expected_values, expected_bool, expected_repr):
    # python -m pytest tests/unit/test_api.py -s -k test_bookmark_initialization_with_values_none
    bookmark = neo4j.api.Bookmark(*test_input)
    assert bookmark.values == expected_values
    assert bool(bookmark) is expected_bool
    assert repr(bookmark) == expected_repr


@pytest.mark.parametrize(
    "test_input, expected_values, expected_bool, expected_repr",
    [
        (("",), frozenset(), False, "<Bookmark values={}>"),
        (("", ""), frozenset(), False, "<Bookmark values={}>"),
        (("bookmark1", ""), frozenset({"bookmark1"}), True, "<Bookmark values={'bookmark1'}>"),
        (("bookmark1", "", "bookmark2", ""), frozenset({"bookmark1", "bookmark2"}), True, "<Bookmark values={'bookmark1', 'bookmark2'}>"),
        (("", "bookmark1", "", "bookmark2", "", "", "bookmark3"), frozenset({"bookmark1", "bookmark2", "bookmark3"}), True, "<Bookmark values={'bookmark1', 'bookmark2', 'bookmark3'}>"),
    ]
)
def test_bookmark_initialization_with_values_empty_string(test_input, expected_values, expected_bool, expected_repr):
    # python -m pytest tests/unit/test_api.py -s -k test_bookmark_initialization_with_values_empty_string
    bookmark = neo4j.api.Bookmark(*test_input)
    assert bookmark.values == expected_values
    assert bool(bookmark) is expected_bool
    assert repr(bookmark) == expected_repr


@pytest.mark.parametrize(
    "test_input, expected_values, expected_bool, expected_repr",
    [
        (("bookmark1",), frozenset({"bookmark1"}), True, "<Bookmark values={'bookmark1'}>"),
        (("bookmark1", "bookmark2", "bookmark3"), frozenset({"bookmark1", "bookmark2", "bookmark3"}), True, "<Bookmark values={'bookmark1', 'bookmark2', 'bookmark3'}>"),
        (standard_ascii, frozenset(standard_ascii), True, "<Bookmark values={{'{values}'}}>".format(values="', '".join(standard_ascii)))
    ]
)
def test_bookmark_initialization_with_valid_strings(test_input, expected_values, expected_bool, expected_repr):
    # python -m pytest tests/unit/test_api.py -s -k test_bookmark_initialization_with_valid_strings
    bookmark = neo4j.api.Bookmark(*test_input)
    assert bookmark.values == expected_values
    assert bool(bookmark) is expected_bool
    assert repr(bookmark) == expected_repr


@pytest.mark.parametrize(
    "test_input, expected",
    [
        ((not_ascii,), ValueError),
        (("", not_ascii,), ValueError),
        (("bookmark1", chr(129),), ValueError),
    ]
)
def test_bookmark_initialization_with_invalid_strings(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_bookmark_initialization_with_invalid_strings
    with pytest.raises(expected) as e:
        bookmark = neo4j.api.Bookmark(*test_input)


@pytest.mark.parametrize(
    "test_input, expected_str, expected_repr",
    [
        ((), "", "Version()"),
        ((None,), "None", "Version(None,)"),
        (("3",), "3", "Version('3',)"),
        (("3", "0"), "3.0", "Version('3', '0')"),
        ((3,), "3", "Version(3,)"),
        ((3, 0), "3.0", "Version(3, 0)"),
        ((3, 0, 0), "3.0.0", "Version(3, 0, 0)"),
        ((3, 0, 0, 0), "3.0.0.0", "Version(3, 0, 0, 0)"),
    ]
)
def test_version_initialization(test_input, expected_str, expected_repr):
    # python -m pytest tests/unit/test_api.py -s -k test_version_initialization
    version = neo4j.api.Version(*test_input)
    assert str(version) == expected_str
    assert repr(version) == expected_repr


@pytest.mark.parametrize(
    "test_input, expected_str, expected_repr",
    [
        (bytearray([0, 0, 0, 0]), "0.0", "Version(0, 0)"),
        (bytearray([0, 0, 0, 1]), "1.0", "Version(1, 0)"),
        (bytearray([0, 0, 1, 0]), "0.1", "Version(0, 1)"),
        (bytearray([0, 0, 1, 1]), "1.1", "Version(1, 1)"),
        (bytearray([0, 0, 254, 254]), "254.254", "Version(254, 254)"),
    ]
)
def test_version_from_bytes_with_valid_bolt_version_handshake(test_input, expected_str, expected_repr):
    # python -m pytest tests/unit/test_api.py -s -k test_version_from_bytes_with_valid_bolt_version_handshake
    version = neo4j.api.Version.from_bytes(test_input)
    assert str(version) == expected_str
    assert repr(version) == expected_repr


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (bytearray([0, 0, 0]), ValueError),
        (bytearray([0, 0, 0, 0, 0]), ValueError),
        (bytearray([1, 0, 0, 0]), ValueError),
        (bytearray([0, 1, 0, 0]), ValueError),
        (bytearray([1, 1, 0, 0]), ValueError),
    ]
)
def test_version_from_bytes_with_not_valid_bolt_version_handshake(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_version_from_bytes_with_not_valid_bolt_version_handshake
    with pytest.raises(expected):
        version = neo4j.api.Version.from_bytes(test_input)


@pytest.mark.parametrize(
    "test_input, expected",
    [
        ((), bytearray([0, 0, 0, 0])),
        ((0,), bytearray([0, 0, 0, 0])),
        ((1,), bytearray([0, 0, 0, 1])),
        ((0, 0), bytearray([0, 0, 0, 0])),
        ((1, 0), bytearray([0, 0, 0, 1])),
        ((1, 2), bytearray([0, 0, 2, 1])),
        ((255, 255), bytearray([0, 0, 255, 255])),
    ]
)
def test_version_to_bytes_with_valid_bolt_version(test_input, expected):
    # python -m pytest tests/unit/test_api.py -s -k test_version_to_bytes_with_valid_bolt_version
    version = neo4j.api.Version(*test_input)
    assert version.to_bytes() == expected


def test_serverinfo_initialization():
    # python -m pytest tests/unit/test_api.py -s -k test_serverinfo_initialization

    from neo4j.addressing import Address

    address = Address(("bolt://localhost", 7687))
    version = neo4j.api.Version(3, 0)

    server_info = neo4j.api.ServerInfo(address, version)
    assert server_info.address is address
    assert server_info.protocol_version is version
    assert server_info.agent is None
    assert server_info.connection_id is None


@pytest.mark.parametrize(
    "test_input, expected_agent, expected_version_info",
    [
        ({"server": "Neo4j/3.0.0"}, "Neo4j/3.0.0", (3, 0, 0)),
        ({"server": "Neo4j/3.X.Y"}, "Neo4j/3.X.Y", (3, "X", "Y")),
    ]
)
def test_serverinfo_with_metadata(test_input, expected_agent, expected_version_info):
    # python -m pytest tests/unit/test_api.py -s -k test_serverinfo_with_metadata
    from neo4j.addressing import Address

    address = Address(("bolt://localhost", 7687))
    version = neo4j.api.Version(3, 0)

    server_info = neo4j.api.ServerInfo(address, version)

    server_info.update(test_input)

    assert server_info.agent == expected_agent
    assert server_info.version_info() == expected_version_info


@pytest.mark.parametrize(
    "test_input, expected_driver_type, expected_security_type, expected_error",
    [
        ("bolt://localhost:7676", neo4j.api.DRIVER_BOLT, neo4j.api.SECURITY_TYPE_NOT_SECURE, None),
        ("bolt+ssc://localhost:7676", neo4j.api.DRIVER_BOLT, neo4j.api.SECURITY_TYPE_SELF_SIGNED_CERTIFICATE, None),
        ("bolt+s://localhost:7676", neo4j.api.DRIVER_BOLT, neo4j.api.SECURITY_TYPE_SECURE, None),
        ("neo4j://localhost:7676", neo4j.api.DRIVER_NEO4j, neo4j.api.SECURITY_TYPE_NOT_SECURE, None),
        ("neo4j+ssc://localhost:7676", neo4j.api.DRIVER_NEO4j, neo4j.api.SECURITY_TYPE_SELF_SIGNED_CERTIFICATE, None),
        ("neo4j+s://localhost:7676", neo4j.api.DRIVER_NEO4j, neo4j.api.SECURITY_TYPE_SECURE, None),
        ("undefined://localhost:7676", None, None, ConfigurationError),
        ("localhost:7676", None, None, ConfigurationError),
        ("://localhost:7676", None, None, ConfigurationError),
        ("bolt+routing://localhost:7676", neo4j.api.DRIVER_NEO4j, neo4j.api.SECURITY_TYPE_NOT_SECURE, ConfigurationError),
        ("bolt://username@localhost:7676", None, None, ConfigurationError),
        ("bolt://username:password@localhost:7676", None, None, ConfigurationError),
    ]
)
def test_uri_scheme(test_input, expected_driver_type, expected_security_type, expected_error):
    # python -m pytest tests/unit/test_api.py -s -k test_uri_scheme
    if expected_error:
        with pytest.raises(expected_error):
            neo4j.api.parse_neo4j_uri(test_input)
    else:
        driver_type, security_type, parsed = neo4j.api.parse_neo4j_uri(test_input)
        assert driver_type == expected_driver_type
        assert security_type == expected_security_type


def test_parse_routing_context():
    # python -m pytest tests/unit/test_api.py -s -v -k test_parse_routing_context
    context = neo4j.api.parse_routing_context(query="name=molly&color=white")
    assert context == {"name": "molly", "color": "white"}


def test_parse_routing_context_should_error_when_value_missing():
    # python -m pytest tests/unit/test_api.py -s -v -k test_parse_routing_context_should_error_when_value_missing
    with pytest.raises(ConfigurationError):
        neo4j.api.parse_routing_context("name=&color=white")


def test_parse_routing_context_should_error_when_key_duplicate():
    # python -m pytest tests/unit/test_api.py -s -v -k test_parse_routing_context_should_error_when_key_duplicate
    with pytest.raises(ConfigurationError):
        neo4j.api.parse_routing_context("name=molly&name=white")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from uuid import uuid4

from neo4j.api import READ_ACCESS, WRITE_ACCESS
from neo4j.graph import Node


def test_can_obtain_bookmark_after_commit(bolt_driver):
    with bolt_driver.session() as session:
        with session.begin_transaction() as tx:
            tx.run("RETURN 1")
        assert session.last_bookmark() is not None


def test_can_pass_bookmark_into_next_transaction(driver):
    unique_id = uuid4().hex

    with driver.session(default_access_mode=WRITE_ACCESS) as session:
        with session.begin_transaction() as tx:
            tx.run("CREATE (a:Thing {uuid:$uuid})", uuid=unique_id)
        bookmark = session.last_bookmark()

    assert bookmark is not None

    with driver.session(default_access_mode=READ_ACCESS, bookmarks=[bookmark]) as session:
        with session.begin_transaction() as tx:
            result = tx.run("MATCH (a:Thing {uuid:$uuid}) RETURN a", uuid=unique_id)
            record_list = list(result)
            assert len(record_list) == 1
            record = record_list[0]
            assert len(record) == 1
            thing = record[0]
            assert isinstance(thing, Node)
            assert thing["uuid"] == unique_id


def test_bookmark_should_be_none_after_rollback(driver):
    with driver.session(default_access_mode=WRITE_ACCESS) as session:
        with session.begin_transaction() as tx:
            tx.run("CREATE (a)")

    assert session.last_bookmark() is not None

    with driver.session(default_access_mode=WRITE_ACCESS) as session:
        with session.begin_transaction() as tx:
            tx.run("CREATE (a)")
            tx.rollback()

    assert session.last_bookmark() is None



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from neo4j.exceptions import ServiceUnavailable
from neo4j._exceptions import BoltHandshakeError

# python -m pytest tests/integration/test_readme.py -s -v


def test_should_run_readme(uri, auth):
    names = set()
    print = names.add

    from neo4j import GraphDatabase

    try:
        driver = GraphDatabase.driver(uri, auth=auth)
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])

    def print_friends(tx, name):
        for record in tx.run("MATCH (a:Person)-[:KNOWS]->(friend) "
                             "WHERE a.name = $name "
                             "RETURN friend.name", name=name):
            print(record["friend.name"])

    with driver.session() as session:
        session.run("MATCH (a) DETACH DELETE a")
        session.run("CREATE (a:Person {name:'Alice'})-[:KNOWS]->({name:'Bob'})")
        session.read_transaction(print_friends, "Alice")

    driver.close()
    assert len(names) == 1
    assert "Bob" in names



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from neo4j import (
    ResultSummary,
    SummaryCounters,
    GraphDatabase,
)
from neo4j.exceptions import (
    ServiceUnavailable,
)
from neo4j._exceptions import (
    BoltHandshakeError,
)


def test_can_obtain_summary_after_consuming_result(session):
    # python -m pytest tests/integration/test_summary.py -s -v -k test_can_obtain_summary_after_consuming_result

    result = session.run("CREATE (n) RETURN n")
    summary = result.consume()
    assert summary.query == "CREATE (n) RETURN n"
    assert summary.parameters == {}
    assert summary.query_type == "rw"
    assert summary.counters.nodes_created == 1


def test_no_plan_info(session):
    result = session.run("CREATE (n) RETURN n")
    summary = result.consume()
    assert summary.plan is None
    assert summary.profile is None


def test_can_obtain_plan_info(session):
    # python -m pytest tests/integration/test_summary.py -s -v -k test_can_obtain_plan_info
    result = session.run("EXPLAIN CREATE (n) RETURN n")
    summary = result.consume()
    assert isinstance(summary.plan, dict)


def test_can_obtain_profile_info(session):
    # python -m pytest tests/integration/test_summary.py -s -v -k test_can_obtain_profile_info
    result = session.run("PROFILE CREATE (n) RETURN n")
    summary = result.consume()
    assert isinstance(summary.profile, dict)


def test_no_notification_info(session):
    # python -m pytest tests/integration/test_summary.py -s -v -k test_no_notification_info
    result = session.run("CREATE (n) RETURN n")
    summary = result.consume()
    assert summary.notifications is None


def test_can_obtain_notification_info(session):
    # python -m pytest tests/integration/test_summary.py -s -v -k test_can_obtain_notification_info
    result = session.run("EXPLAIN MATCH (n), (m) RETURN n, m")
    summary = result.consume()
    assert isinstance(summary, ResultSummary)

    notifications = summary.notifications
    assert isinstance(notifications, list)
    assert len(notifications) == 1
    assert isinstance(notifications[0], dict)


def test_contains_time_information(session):
    summary = session.run("UNWIND range(1,1000) AS n RETURN n AS number").consume()

    assert isinstance(summary.result_available_after, int)
    assert isinstance(summary.result_consumed_after, int)

    with pytest.raises(AttributeError) as ex:
        assert isinstance(summary.t_first, int)

    with pytest.raises(AttributeError) as ex:
        assert isinstance(summary.t_last, int)


def test_protocol_version_information(session):
    summary = session.run("UNWIND range(1,100) AS n RETURN n AS number").consume()

    assert isinstance(summary.server.protocol_version, tuple)
    assert isinstance(summary.server.protocol_version[0], int)
    assert isinstance(summary.server.protocol_version[1], int)


def test_summary_counters_case_1(session):
    # python -m pytest tests/integration/test_summary.py -s -v -k test_summary_counters_case_1

    result = session.run("RETURN $number AS x", number=3)
    summary = result.consume()

    assert summary.query == "RETURN $number AS x"
    assert summary.parameters == {"number": 3}

    assert isinstance(summary.query_type, str)

    counters = summary.counters

    assert isinstance(counters, SummaryCounters)
    assert counters.nodes_created == 0
    assert counters.nodes_deleted == 0
    assert counters.relationships_created == 0
    assert counters.relationships_deleted == 0
    assert counters.properties_set == 0
    assert counters.labels_added == 0
    assert counters.labels_removed == 0
    assert counters.indexes_added == 0
    assert counters.indexes_removed == 0
    assert counters.constraints_added == 0
    assert counters.constraints_removed == 0
    assert counters.contains_updates is False

    assert counters.system_updates == 0
    assert counters.contains_system_updates is False


def test_summary_counters_case_2(neo4j_uri, auth, target, requires_bolt_4x):
    # python -m pytest tests/integration/test_summary.py -s -v -k test_summary_counters_case_2
    with GraphDatabase.driver(neo4j_uri, auth=auth) as driver:

        with driver.session(database="system") as session:
            session.run("DROP DATABASE test IF EXISTS").consume()

            # SHOW DATABASES

            result = session.run("SHOW DATABASES")
            databases = set()
            for record in result:
                databases.add(record.get("name"))
            assert "system" in databases
            assert "neo4j" in databases

            summary = result.consume()

            assert summary.query == "SHOW DATABASES"
            assert summary.parameters == {}

            assert isinstance(summary.query_type, str)

            counters = summary.counters

            assert isinstance(counters, SummaryCounters)
            assert counters.nodes_created == 0
            assert counters.nodes_deleted == 0
            assert counters.relationships_created == 0
            assert counters.relationships_deleted == 0
            assert counters.properties_set == 0
            assert counters.labels_added == 0
            assert counters.labels_removed == 0
            assert counters.indexes_added == 0
            assert counters.indexes_removed == 0
            assert counters.constraints_added == 0
            assert counters.constraints_removed == 0
            assert counters.contains_updates is False

            assert counters.system_updates == 0
            assert counters.contains_system_updates is False

            # CREATE DATABASE test

            summary = session.run("CREATE DATABASE test").consume()

            assert summary.query == "CREATE DATABASE test"
            assert summary.parameters == {}

            assert isinstance(summary.query_type, str)

            counters = summary.counters

            assert isinstance(counters, SummaryCounters)
            assert counters.nodes_created == 0
            assert counters.nodes_deleted == 0
            assert counters.relationships_created == 0
            assert counters.relationships_deleted == 0
            assert counters.properties_set == 0
            assert counters.labels_added == 0
            assert counters.labels_removed == 0
            assert counters.indexes_added == 0
            assert counters.indexes_removed == 0
            assert counters.constraints_added == 0
            assert counters.constraints_removed == 0
            assert counters.contains_updates is False

            assert counters.system_updates == 1
            assert counters.contains_system_updates is True

        with driver.session(database="test") as session:
            summary = session.run("CREATE (n)").consume()
            assert summary.counters.contains_updates is True
            assert summary.counters.contains_system_updates is False
            assert summary.counters.nodes_created == 1

        with driver.session(database="test") as session:
            summary = session.run("MATCH (n) DELETE (n)").consume()
            assert summary.counters.contains_updates is True
            assert summary.counters.contains_system_updates is False
            assert summary.counters.nodes_deleted == 1

        with driver.session(database="test") as session:
            summary = session.run("CREATE ()-[:KNOWS]->()").consume()
            assert summary.counters.contains_updates is True
            assert summary.counters.contains_system_updates is False
            assert summary.counters.relationships_created == 1

        with driver.session(database="test") as session:
            summary = session.run("MATCH ()-[r:KNOWS]->() DELETE r").consume()
            assert summary.counters.contains_updates is True
            assert summary.counters.contains_system_updates is False
            assert summary.counters.relationships_deleted == 1

        with driver.session(database="test") as session:
            summary = session.run("CREATE (n:ALabel)").consume()
            assert summary.counters.contains_updates is True
            assert summary.counters.contains_system_updates is False
            assert summary.counters.labels_added == 1

        with driver.session(database="test") as session:
            summary = session.run("MATCH (n:ALabel) REMOVE n:ALabel").consume()
            assert summary.counters.contains_updates is True
            assert summary.counters.contains_system_updates is False
            assert summary.counters.labels_removed == 1

        with driver.session(database="test") as session:
            summary = session.run("CREATE (n {magic: 42})").consume()
            assert summary.counters.contains_updates is True
            assert summary.counters.contains_system_updates is False
            assert summary.counters.properties_set == 1

        with driver.session(database="test") as session:
            summary = session.run("CREATE INDEX ON :ALabel(prop)").consume()
            assert summary.counters.contains_updates is True
            assert summary.counters.contains_system_updates is False
            assert summary.counters.indexes_added == 1

        with driver.session(database="test") as session:
            summary = session.run("DROP INDEX ON :ALabel(prop)").consume()
            assert summary.counters.contains_updates is True
            assert summary.counters.contains_system_updates is False
            assert summary.counters.indexes_removed == 1

        with driver.session(database="test") as session:
            summary = session.run("CREATE CONSTRAINT ON (book:Book) ASSERT book.isbn IS UNIQUE").consume()
            assert summary.counters.contains_updates is True
            assert summary.counters.contains_system_updates is False
            assert summary.counters.constraints_added == 1

        with driver.session(database="test") as session:
            summary = session.run("DROP CONSTRAINT ON (book:Book) ASSERT book.isbn IS UNIQUE").consume()
            assert summary.counters.contains_updates is True
            assert summary.counters.contains_system_updates is False
            assert summary.counters.constraints_removed == 1

        with driver.session(database="system") as session:
            session.run("DROP DATABASE test IF EXISTS").consume()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from uuid import uuid4


from neo4j.packstream import Structure
from neo4j.exceptions import (
    Neo4jError,
    CypherSyntaxError,
)
from neo4j.graph import (
    Node,
    Relationship,
    Path,
)
from neo4j.work.pipelining import PullOrderException


def test_can_run_simple_statement(bolt_driver):
    pipeline = bolt_driver.pipeline(flush_every=0)
    pipeline.push("RETURN 1 AS n")
    for record in pipeline.pull():
        assert len(record) == 1
        assert record[0] == 1
        # TODO: why does pipeline result not look like a regular result?
        # assert record["n"] == 1
        # with pytest.raises(KeyError):
        #    _ = record["x"]
        # assert record["n"] == 1
        # with pytest.raises(KeyError):
        #    _ = record["x"]
        with pytest.raises(TypeError):
            _ = record[object()]
        assert repr(record)
        assert len(record) == 1
    pipeline.close()


def test_can_run_simple_statement_with_params(bolt_driver):
    pipeline = bolt_driver.pipeline(flush_every=0)
    count = 0
    pipeline.push("RETURN $x AS n", {"x": {"abc": ["d", "e", "f"]}})
    for record in pipeline.pull():
        assert record[0] == {"abc": ["d", "e", "f"]}
        # TODO: why does pipeline result not look like a regular result?
        # assert record["n"] == {"abc": ["d", "e", "f"]}
        assert repr(record)
        assert len(record) == 1
        count += 1
    pipeline.close()
    assert count == 1


def test_can_run_write_statement_with_no_return(driver):
    pipeline = driver.pipeline(flush_every=0)
    count = 0
    test_uid = str(uuid4())
    pipeline.push("CREATE (a:Person {uid:$test_uid})", dict(test_uid=test_uid))

    for _ in pipeline.pull():
        raise Exception("Should not return any results from create with no return")
    # Note you still have to consume the generator if you want to be allowed to pull from the pipeline again even
    # though it doesn't apparently return any items.

    pipeline.push("MATCH (a:Person {uid:$test_uid}) RETURN a LIMIT 1", dict(test_uid=test_uid))
    for _ in pipeline.pull():
        count += 1
    pipeline.close()
    assert count == 1


def test_fails_on_bad_syntax(bolt_driver):
    pipeline = bolt_driver.pipeline(flush_every=0)
    with pytest.raises(Neo4jError):
        pipeline.push("X")
        next(pipeline.pull())


def test_doesnt_fail_on_bad_syntax_somewhere(bolt_driver):
    pipeline = bolt_driver.pipeline(flush_every=0)
    pipeline.push("RETURN 1 AS n")
    pipeline.push("X")
    assert next(pipeline.pull())[0] == 1
    with pytest.raises(Neo4jError):
        next(pipeline.pull())


def test_fails_on_missing_parameter(bolt_driver):
    pipeline = bolt_driver.pipeline(flush_every=0)
    with pytest.raises(Neo4jError):
        pipeline.push("RETURN $x")
        next(pipeline.pull())


def test_can_run_simple_statement_from_bytes_string(bolt_driver):
    pipeline = bolt_driver.pipeline(flush_every=0)
    count = 0
    pytest.skip("FIXME: why can't pipeline handle bytes string?")
    pipeline.push(b"RETURN 1 AS n")
    for record in pipeline.pull():
        assert record[0] == 1
        assert record["n"] == 1
        assert repr(record)
        assert len(record) == 1
        count += 1
    pipeline.close()
    assert count == 1


def test_can_run_statement_that_returns_multiple_records(bolt_driver):
    pipeline = bolt_driver.pipeline(flush_every=0)
    count = 0
    pipeline.push("unwind(range(1, 10)) AS z RETURN z")
    for record in pipeline.pull():
        assert 1 <= record[0] <= 10
        count += 1
    pipeline.close()
    assert count == 10


def test_can_return_node(neo4j_driver):
    with neo4j_driver.pipeline(flush_every=0) as pipeline:
        pipeline.push("CREATE (a:Person {name:'Alice'}) RETURN a")
        record_list = list(pipeline.pull())
        assert len(record_list) == 1
        for record in record_list:
            alice = record[0]
            print(alice)
            pytest.skip("FIXME: why does pipeline result not look like a regular result?")
            assert isinstance(alice, Node)
            assert alice.labels == {"Person"}
            assert dict(alice) == {"name": "Alice"}


def test_can_return_relationship(neo4j_driver):
    with neo4j_driver.pipeline(flush_every=0) as pipeline:
        pipeline.push("CREATE ()-[r:KNOWS {since:1999}]->() RETURN r")
        record_list = list(pipeline.pull())
        assert len(record_list) == 1
        for record in record_list:
            rel = record[0]
            print(rel)
            pytest.skip("FIXME: why does pipeline result not look like a regular result?")
            assert isinstance(rel, Relationship)
            assert rel.type == "KNOWS"
            assert dict(rel) == {"since": 1999}


def test_can_return_path(neo4j_driver):
    with neo4j_driver.pipeline(flush_every=0) as pipeline:
        test_uid = str(uuid4())
        pipeline.push(
            "MERGE p=(alice:Person {name:'Alice', test_uid: $test_uid})"
            "-[:KNOWS {test_uid: $test_uid}]->"
            "(:Person {name:'Bob', test_uid: $test_uid})"
            " RETURN p",
            dict(test_uid=test_uid)
        )
        record_list = list(pipeline.pull())
        assert len(record_list) == 1
        for record in record_list:
            path = record[0]
            print(path)
            pytest.skip("FIXME: why does pipeline result not look like a regular result?")
            assert isinstance(path, Path)
            assert path.start_node["name"] == "Alice"
            assert path.end_node["name"] == "Bob"
            assert path.relationships[0].type == "KNOWS"
            assert len(path.nodes) == 2
            assert len(path.relationships) == 1


def test_can_handle_cypher_error(bolt_driver):
    with bolt_driver.pipeline(flush_every=0) as pipeline:
        pipeline.push("X")
        with pytest.raises(Neo4jError):
            next(pipeline.pull())


def test_should_not_allow_empty_statements(bolt_driver):
    with bolt_driver.pipeline(flush_every=0) as pipeline:
        pipeline.push("")
        with pytest.raises(CypherSyntaxError):
            next(pipeline.pull())


def test_can_queue_multiple_statements(bolt_driver):
    count = 0
    with bolt_driver.pipeline(flush_every=0) as pipeline:
        pipeline.push("unwind(range(1, 10)) AS z RETURN z")
        pipeline.push("unwind(range(11, 20)) AS z RETURN z")
        pipeline.push("unwind(range(21, 30)) AS z RETURN z")
        for i in range(3):
            for record in pipeline.pull():
                assert (i * 10 + 1) <= record[0] <= ((i + 1) * 10)
                count += 1
    assert count == 30


def test_pull_order_exception(bolt_driver):
    """If you try and pull when you haven't finished iterating the previous result you get an error"""
    pipeline = bolt_driver.pipeline(flush_every=0)
    with pytest.raises(PullOrderException):
        pipeline.push("unwind(range(1, 10)) AS z RETURN z")
        pipeline.push("unwind(range(11, 20)) AS z RETURN z")
        generator_one = pipeline.pull()
        generator_two = pipeline.pull()


def test_pipeline_can_read_own_writes(neo4j_driver):
    """I am not sure that we _should_ guarantee this"""
    count = 0
    with neo4j_driver.pipeline(flush_every=0) as pipeline:
        test_uid = str(uuid4())
        pipeline.push(
            "CREATE (a:Person {name:'Alice', test_uid: $test_uid})",
            dict(test_uid=test_uid)
        )
        pipeline.push(
            "MATCH (alice:Person {name:'Alice', test_uid: $test_uid}) "
            "MERGE (alice)"
            "-[:KNOWS {test_uid: $test_uid}]->"
            "(:Person {name:'Bob', test_uid: $test_uid})",
            dict(test_uid=test_uid)
        )
        pipeline.push("MATCH (n:Person {test_uid: $test_uid}) RETURN n", dict(test_uid=test_uid))
        pipeline.push(
            "MATCH"
            " p=(:Person {test_uid: $test_uid})-[:KNOWS {test_uid: $test_uid}]->(:Person {test_uid: $test_uid})"
            " RETURN p",
            dict(test_uid=test_uid)
        )

        # create Alice
        # n.b. we have to consume the result
        assert next(pipeline.pull(), True) == True

        # merge "knows Bob"
        # n.b. we have to consume the result
        assert next(pipeline.pull(), True) == True

        # get people
        for result in pipeline.pull():
            count += 1

            assert len(result) == 1
            person = result[0]
            print(person)
            assert isinstance(person, Structure)
            assert person.tag == b'N'
            print(person.fields)
            assert set(person.fields[1]) == {"Person"}

        print(count)
        assert count == 2

        # get path
        for result in pipeline.pull():
            count += 1

            assert len(result) == 1
            path = result[0]
            print(path)
            assert isinstance(path, Structure)
            assert path.tag == b'P'

            # TODO: return Path / Node / Rel instances rather than Structures
            # assert isinstance(path, Path)
            # assert path.start_node["name"] == "Alice"
            # assert path.end_node["name"] == "Bob"
            # assert path.relationships[0].type == "KNOWS"
            # assert len(path.nodes) == 2
            # assert len(path.relationships) == 1

    assert count == 3


def test_automatic_reset_after_failure(bolt_driver):
    with bolt_driver.pipeline(flush_every=0) as pipeline:
        try:
            pipeline.push("X")
            next(pipeline.pull())
        except Neo4jError:
            pipeline.push("RETURN 1")
            record = next(pipeline.pull())
            assert record[0] == 1
        else:
            assert False, "A Cypher error should have occurred"



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unittest import TestCase

from neo4j.data import DataHydrator
from neo4j.packstream import Structure


class TemporalHydrationTestCase(TestCase):

    def setUp(self):
        self.hydrant = DataHydrator()

    def test_can_hydrate_date_time_structure(self):
        struct = Structure(b'd', 1539344261, 474716862)
        dt, = self.hydrant.hydrate([struct])
        self.assertEqual(dt.year, 2018)
        self.assertEqual(dt.month, 10)
        self.assertEqual(dt.day, 12)
        self.assertEqual(dt.hour, 11)
        self.assertEqual(dt.minute, 37)
        self.assertEqual(dt.second, 41.474716862)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from contextlib import redirect_stdout
from io import StringIO

# tag::driver-introduction-example-import[]
from neo4j import GraphDatabase
import logging
from neo4j.exceptions import ServiceUnavailable
# end::driver-introduction-example-import[]

from neo4j._exceptions import BoltHandshakeError


# python -m pytest tests/integration/examples/test_driver_introduction_example.py -s -v

# tag::driver-introduction-example[]
class App:

    def __init__(self, uri, user, password):
        self.driver = GraphDatabase.driver(uri, auth=(user, password))

    def close(self):
        # Don't forget to close the driver connection when you are finished with it
        self.driver.close()

    def create_friendship(self, person1_name, person2_name):
        with self.driver.session() as session:
            # Write transactions allow the driver to handle retries and transient errors
            result = session.write_transaction(
                self._create_and_return_friendship, person1_name, person2_name)
            for row in result:
                print("Created friendship between: {p1}, {p2}".format(p1=row['p1'], p2=row['p2']))

    @staticmethod
    def _create_and_return_friendship(tx, person1_name, person2_name):
        # To learn more about the Cypher syntax, see https://neo4j.com/docs/cypher-manual/current/
        # The Reference Card is also a good resource for keywords https://neo4j.com/docs/cypher-refcard/current/
        query = (
            "CREATE (p1:Person { name: $person1_name }) "
            "CREATE (p2:Person { name: $person2_name }) "
            "CREATE (p1)-[:KNOWS]->(p2) "
            "RETURN p1, p2"
        )
        result = tx.run(query, person1_name=person1_name, person2_name=person2_name)
        try:
            return [{"p1": row["p1"]["name"], "p2": row["p2"]["name"]}
                    for row in result]
        # Capture any errors along with the query and data for traceability
        except ServiceUnavailable as exception:
            logging.error("{query} raised an error: \n {exception}".format(
                query=query, exception=exception))
            raise

    def find_person(self, person_name):
        with self.driver.session() as session:
            result = session.read_transaction(self._find_and_return_person, person_name)
            for row in result:
                print("Found person: {row}".format(row=row))

    @staticmethod
    def _find_and_return_person(tx, person_name):
        query = (
            "MATCH (p:Person) "
            "WHERE p.name = $person_name "
            "RETURN p.name AS name"
        )
        result = tx.run(query, person_name=person_name)
        return [row["name"] for row in result]


if __name__ == "__main__":
    bolt_url = "%%BOLT_URL_PLACEHOLDER%%"
    user = "<Username for database>"
    password = "<Password for database>"
    app = App(bolt_url, user, password)
    app.create_friendship("Alice", "David")
    app.find_person("Alice")
    app.close()
# end::driver-introduction-example[]


def test_driver_introduction_example(uri, auth):
    try:
        s = StringIO()
        with redirect_stdout(s):
            app = App(uri, auth[0], auth[1])
            app.create_friendship("Alice", "David")
            app.find_person("Alice")
            app.close()

        # assert s.getvalue().startswith("Found person: Alice")
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j import (
    GraphDatabase,
    Neo4jDriver,
    Version,
    READ_ACCESS,
    ResultSummary,
    ServerInfo,
)
from neo4j.exceptions import (
    ServiceUnavailable,
    ConfigurationError,
    ClientError,
)
from neo4j._exceptions import (
    BoltHandshakeError,
)
from neo4j.conf import (
    RoutingConfig,
)
from neo4j.io._bolt3 import Bolt3

# python -m pytest tests/integration/test_neo4j_driver.py -s -v


def test_neo4j_uri(neo4j_uri, auth, target):
    # python -m pytest tests/integration/test_neo4j_driver.py -s -v -k test_neo4j_uri
    try:
        with GraphDatabase.driver(neo4j_uri, auth=auth) as driver:
            with driver.session() as session:
                value = session.run("RETURN 1").single().value()
                assert value == 1
    except ServiceUnavailable as error:
        if error.args[0] == "Server does not support routing":
            # This is because a single instance Neo4j 3.5 does not have dbms.routing.cluster.getRoutingTable() call
            pytest.skip(error.args[0])
        elif isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


def test_supports_multi_db(neo4j_driver, requires_bolt_4x):
    # python -m pytest tests/integration/test_neo4j_driver.py -s -v -k test_supports_multi_db
    assert neo4j_driver.supports_multi_db()


def test_test_multi_db_specify_database(neo4j_uri, auth, target):
    # python -m pytest tests/integration/test_neo4j_driver.py -s -v -k test_test_multi_db_specify_database
    try:
        with GraphDatabase.driver(neo4j_uri, auth=auth, database="test_database") as driver:
            with driver.session() as session:
                result = session.run("RETURN 1")
                assert next(result) == 1
                summary = result.consume()
                assert summary.database == "test_database"
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])
    except ConfigurationError as error:
        assert "Database name parameter for selecting database is not supported in Bolt Protocol Version(3, 0)." in error.args[0]
    except ClientError as error:
        # FAILURE {'code': 'Neo.ClientError.Database.DatabaseNotFound' - This message is sent from the server
        assert error.code == "Neo.ClientError.Database.DatabaseNotFound"
        assert "test_database" in error.message


def test_neo4j_multi_database_support_create(neo4j_uri, auth, target, requires_bolt_4x):
    # python -m pytest tests/integration/test_neo4j_driver.py -s -v -k test_neo4j_multi_database_support_create
    with GraphDatabase.driver(neo4j_uri, auth=auth) as driver:
        with driver.session(database="system") as session:
            session.run("DROP DATABASE test IF EXISTS").consume()
            result = session.run("SHOW DATABASES")
            databases = set()
            for record in result:
                databases.add(record.get("name"))
            assert "system" in databases
            assert "neo4j" in databases

            session.run("CREATE DATABASE test").consume()
            result = session.run("SHOW DATABASES")
            for record in result:
                databases.add(record.get("name"))
            assert "system" in databases
            assert "neo4j" in databases
            assert "test" in databases
        with driver.session(database="system") as session:
            session.run("DROP DATABASE test IF EXISTS").consume()


def test_neo4j_multi_database_support_different(neo4j_uri, auth, target, requires_bolt_4x):
    # python -m pytest tests/integration/test_neo4j_driver.py -s -v -k test_neo4j_multi_database_support_different
    with GraphDatabase.driver(neo4j_uri, auth=auth) as driver:
        with driver.session() as session:
            # Test that default database is empty
            session.run("MATCH (n) DETACH DELETE n").consume()
            result = session.run("MATCH (p:Person) RETURN p")
            names = set()
            for ix in result:
                names.add(ix["p"].get("name"))
            assert names == set()  # THIS FAILS?
        with driver.session(database="system") as session:
            session.run("DROP DATABASE testa IF EXISTS").consume()
            session.run("DROP DATABASE testb IF EXISTS").consume()
        with driver.session(database="system") as session:
            result = session.run("SHOW DATABASES")
            databases = set()
            for record in result:
                databases.add(record.get("name"))
            assert databases == {"system", "neo4j"}
            result = session.run("CREATE DATABASE testa")
            result.consume()
            result = session.run("CREATE DATABASE testb")
            result.consume()
        with driver.session(database="testa") as session:
            result = session.run('CREATE (p:Person {name: "ALICE"})')
            result.consume()
        with driver.session(database="testb") as session:
            result = session.run('CREATE (p:Person {name: "BOB"})')
            result.consume()
        with driver.session() as session:
            # Test that default database is still empty
            result = session.run("MATCH (p:Person) RETURN p")
            names = set()
            for ix in result:
                names.add(ix["p"].get("name"))
            assert names == set()  # THIS FAILS?
        with driver.session(database="testa") as session:
            result = session.run("MATCH (p:Person) RETURN p")
            names = set()
            for ix in result:
                names.add(ix["p"].get("name"))
            assert names == {"ALICE", }
        with driver.session(database="testb") as session:
            result = session.run("MATCH (p:Person) RETURN p")
            names = set()
            for ix in result:
                names.add(ix["p"].get("name"))
            assert names == {"BOB", }
        with driver.session(database="system") as session:
            session.run("DROP DATABASE testa IF EXISTS").consume()
        with driver.session(database="system") as session:
            session.run("DROP DATABASE testb IF EXISTS").consume()
        with driver.session() as session:
            session.run("MATCH (n) DETACH DELETE n").consume()


def test_neo4j_multi_database_test_routing_table_creates_new_if_deleted(neo4j_uri, auth, target, requires_bolt_4x):
    # python -m pytest tests/integration/test_neo4j_driver.py -s -v -k test_neo4j_multi_database_test_routing_table_creates_new_if_deleted
    with GraphDatabase.driver(neo4j_uri, auth=auth) as driver:
        with driver.session(database="system") as session:
            result = session.run("DROP DATABASE test IF EXISTS")
            result.consume()
            result = session.run("SHOW DATABASES")
            databases = set()
            for record in result:
                databases.add(record.get("name"))
            assert databases == {"system", "neo4j"}

            result = session.run("CREATE DATABASE test")
            result.consume()
            result = session.run("SHOW DATABASES")
            for record in result:
                databases.add(record.get("name"))
            assert databases == {"system", "neo4j", "test"}
        with driver.session(database="test") as session:
            result = session.run("RETURN 1 AS x")
            result.consume()
        del driver._pool.routing_tables["test"]
        with driver.session(database="test") as session:
            result = session.run("RETURN 1 AS x")
            result.consume()
        with driver.session(database="system") as session:
            result = session.run("DROP DATABASE test IF EXISTS")
            result.consume()


def test_neo4j_multi_database_test_routing_table_updates_if_stale(neo4j_uri, auth, target, requires_bolt_4x):
    # python -m pytest tests/integration/test_neo4j_driver.py -s -v -k test_neo4j_multi_database_test_routing_table_updates_if_stale
    with GraphDatabase.driver(neo4j_uri, auth=auth) as driver:
        with driver.session(database="system") as session:
            result = session.run("DROP DATABASE test IF EXISTS")
            result.consume()
            result = session.run("SHOW DATABASES")
            databases = set()
            for record in result:
                databases.add(record.get("name"))
            assert databases == {"system", "neo4j"}

            result = session.run("CREATE DATABASE test")
            result.consume()
            result = session.run("SHOW DATABASES")
            for record in result:
                databases.add(record.get("name"))
            assert databases == {"system", "neo4j", "test"}
        with driver.session(database="test") as session:
            result = session.run("RETURN 1 AS x")
            result.consume()
        driver._pool.routing_tables["test"].ttl = 0
        old_value = driver._pool.routing_tables["test"].last_updated_time
        with driver.session(database="test") as session:
            result = session.run("RETURN 1 AS x")
            result.consume()
        with driver.session(database="system") as session:
            result = session.run("DROP DATABASE test IF EXISTS")
            result.consume()
        assert driver._pool.routing_tables["test"].last_updated_time > old_value


def test_neo4j_multi_database_test_routing_table_removes_aged(neo4j_uri, auth, target, requires_bolt_4x):
    # python -m pytest tests/integration/test_neo4j_driver.py -s -v -k test_neo4j_multi_database_test_routing_table_removes_aged
    with GraphDatabase.driver(neo4j_uri, auth=auth) as driver:
        with driver.session(database="system") as session:
            result = session.run("DROP DATABASE testa IF EXISTS")
            result.consume()
            result = session.run("DROP DATABASE testb IF EXISTS")
            result.consume()
            result = session.run("SHOW DATABASES")
            databases = set()
            for record in result:
                databases.add(record.get("name"))
            assert databases == {"system", "neo4j"}

            result = session.run("CREATE DATABASE testa")
            result.consume()
            result = session.run("CREATE DATABASE testb")
            result.consume()
            result = session.run("SHOW DATABASES")
            for record in result:
                databases.add(record.get("name"))
            assert databases == {"system", "neo4j", "testa", "testb"}
        with driver.session(database="testa") as session:
            result = session.run("RETURN 1 AS x")
            result.consume()
        with driver.session(database="testb") as session:
            result = session.run("RETURN 1 AS x")
            result.consume()
        driver._pool.routing_tables["testa"].ttl = 0
        driver._pool.routing_tables["testb"].ttl = -1 * RoutingConfig.routing_table_purge_delay
        old_value = driver._pool.routing_tables["testa"].last_updated_time
        with driver.session(database="testa") as session:
            # This will refresh the routing table for "testa" and the refresh will trigger a cleanup of aged routing tables
            result = session.run("RETURN 1 AS x")
            result.consume()
        with driver.session(database="system") as session:
            result = session.run("DROP DATABASE testa IF EXISTS")
            result.consume()
            result = session.run("DROP DATABASE testb IF EXISTS")
            result.consume()
        assert driver._pool.routing_tables["testa"].last_updated_time > old_value
        assert "testb" not in driver._pool.routing_tables


def test_neo4j_driver_fetch_size_config_autocommit_normal_case(neo4j_uri, auth):
    # python -m pytest tests/integration/test_neo4j_driver.py -s -v -k test_neo4j_driver_fetch_size_config_autocommit_normal_case
    try:
        with GraphDatabase.driver(neo4j_uri, auth=auth, user_agent="test") as driver:
            assert isinstance(driver, Neo4jDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = []
                result = session.run("UNWIND [1,2,3,4] AS x RETURN x")
                for record in result:
                    expected.append(record["x"])

        assert expected == [1, 2, 3, 4]
    except ServiceUnavailable as error:
        if error.args[0] == "Server does not support routing":
            # This is because a single instance Neo4j 3.5 does not have dbms.routing.cluster.getRoutingTable() call
            pytest.skip(error.args[0])
        elif isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


def test_neo4j_driver_fetch_size_config_autocommit_consume_case(neo4j_uri, auth):
    # python -m pytest tests/integration/test_neo4j_driver.py -s -v -k test_neo4j_driver_fetch_size_config_autocommit_consume_case
    try:
        with GraphDatabase.driver(neo4j_uri, auth=auth, user_agent="test") as driver:
            assert isinstance(driver, Neo4jDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                result = session.run("UNWIND [1,2,3,4] AS x RETURN x")
                result_summary_consume = result.consume()

        assert isinstance(result_summary_consume, ResultSummary)
    except ServiceUnavailable as error:
        if error.args[0] == "Server does not support routing":
            # This is because a single instance Neo4j 3.5 does not have dbms.routing.cluster.getRoutingTable() call
            pytest.skip(error.args[0])
        elif isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


def test_neo4j_driver_fetch_size_config_explicit_transaction(neo4j_uri, auth):
    # python -m pytest tests/integration/test_neo4j_driver.py -s -v -k test_neo4j_driver_fetch_size_config_explicit_transaction
    try:
        with GraphDatabase.driver(neo4j_uri, auth=auth, user_agent="test") as driver:
            assert isinstance(driver, Neo4jDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = []
                tx = session.begin_transaction()
                result = tx.run("UNWIND [1,2,3,4] AS x RETURN x")
                for record in result:
                    expected.append(record["x"])
                tx.commit()

        assert expected == [1, 2, 3, 4]
    except ServiceUnavailable as error:
        if error.args[0] == "Server does not support routing":
            # This is because a single instance Neo4j 3.5 does not have dbms.routing.cluster.getRoutingTable() call
            pytest.skip(error.args[0])
        elif isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j.data import Record

# python -m pytest -s -v tests/unit/test_record.py


def test_record_equality():
    record1 = Record(zip(["name", "empire"], ["Nigel", "The British Empire"]))
    record2 = Record(zip(["name", "empire"], ["Nigel", "The British Empire"]))
    record3 = Record(zip(["name", "empire"], ["Stefan", "Das Deutschland"]))
    assert record1 == record2
    assert record1 != record3
    assert record2 != record3


def test_record_hashing():
    record1 = Record(zip(["name", "empire"], ["Nigel", "The British Empire"]))
    record2 = Record(zip(["name", "empire"], ["Nigel", "The British Empire"]))
    record3 = Record(zip(["name", "empire"], ["Stefan", "Das Deutschland"]))
    assert hash(record1) == hash(record2)
    assert hash(record1) != hash(record3)
    assert hash(record2) != hash(record3)


def test_record_iter():
    a_record = Record(zip(["name", "empire"], ["Nigel", "The British Empire"]))
    assert list(a_record.__iter__()) == ["Nigel", "The British Empire"]


def test_record_as_dict():
    a_record = Record(zip(["name", "empire"], ["Nigel", "The British Empire"]))
    assert dict(a_record) == {"name": "Nigel", "empire": "The British Empire"}


def test_record_as_list():
    a_record = Record(zip(["name", "empire"], ["Nigel", "The British Empire"]))
    assert list(a_record) == ["Nigel", "The British Empire"]


def test_record_len():
    a_record = Record(zip(["name", "empire"], ["Nigel", "The British Empire"]))
    assert len(a_record) == 2


def test_record_repr():
    a_record = Record(zip(["name", "empire"], ["Nigel", "The British Empire"]))
    assert repr(a_record) == "<Record name='Nigel' empire='The British Empire'>"


def test_record_data():
    r = Record(zip(["name", "age", "married"], ["Alice", 33, True]))
    assert r.data() == {"name": "Alice", "age": 33, "married": True}
    assert r.data("name") == {"name": "Alice"}
    assert r.data("age", "name") == {"age": 33, "name": "Alice"}
    assert r.data("age", "name", "shoe size") == {"age": 33, "name": "Alice", "shoe size": None}
    assert r.data(0, "name") == {"name": "Alice"}
    assert r.data(0) == {"name": "Alice"}
    assert r.data(1, 0) == {"age": 33, "name": "Alice"}
    with pytest.raises(IndexError):
        _ = r.data(1, 0, 999)


def test_record_keys():
    r = Record(zip(["name", "age", "married"], ["Alice", 33, True]))
    assert r.keys() == ["name", "age", "married"]


def test_record_values():
    r = Record(zip(["name", "age", "married"], ["Alice", 33, True]))
    assert r.values() == ["Alice", 33, True]
    assert r.values("name") == ["Alice"]
    assert r.values("age", "name") == [33, "Alice"]
    assert r.values("age", "name", "shoe size") == [33, "Alice", None]
    assert r.values(0, "name") == ["Alice", "Alice"]
    assert r.values(0) == ["Alice"]
    assert r.values(1, 0) == [33, "Alice"]
    with pytest.raises(IndexError):
        _ = r.values(1, 0, 999)


def test_record_items():
    r = Record(zip(["name", "age", "married"], ["Alice", 33, True]))
    assert r.items() == [("name", "Alice"), ("age", 33), ("married", True)]
    assert r.items("name") == [("name", "Alice")]
    assert r.items("age", "name") == [("age", 33), ("name", "Alice")]
    assert r.items("age", "name", "shoe size") == [("age", 33), ("name", "Alice"), ("shoe size", None)]
    assert r.items(0, "name") == [("name", "Alice"), ("name", "Alice")]
    assert r.items(0) == [("name", "Alice")]
    assert r.items(1, 0) == [("age", 33), ("name", "Alice")]
    with pytest.raises(IndexError):
        _ = r.items(1, 0, 999)


def test_record_index():
    r = Record(zip(["name", "age", "married"], ["Alice", 33, True]))
    assert r.index("name") == 0
    assert r.index("age") == 1
    assert r.index("married") == 2
    with pytest.raises(KeyError):
        _ = r.index("shoe size")
    assert r.index(0) == 0
    assert r.index(1) == 1
    assert r.index(2) == 2
    with pytest.raises(IndexError):
        _ = r.index(3)
    with pytest.raises(TypeError):
        _ = r.index(None)


def test_record_value():
    r = Record(zip(["name", "age", "married"], ["Alice", 33, True]))
    assert r.value() == "Alice"
    assert r.value("name") == "Alice"
    assert r.value("age") == 33
    assert r.value("married") is True
    assert r.value("shoe size") is None
    assert r.value("shoe size", 6) == 6
    assert r.value(0) == "Alice"
    assert r.value(1) == 33
    assert r.value(2) is True
    assert r.value(3) is None
    assert r.value(3, 6) == 6
    with pytest.raises(TypeError):
        _ = r.value(None)


def test_record_contains():
    r = Record(zip(["name", "age", "married"], ["Alice", 33, True]))
    assert "Alice" in r
    assert 33 in r
    assert True in r
    assert 7.5 not in r
    with pytest.raises(TypeError):
        _ = r.index(None)


def test_record_from_dict():
    r = Record({"name": "Alice", "age": 33})
    assert r["name"] == "Alice"
    assert r["age"] == 33


def test_record_get_slice():
    r = Record(zip(["name", "age", "married"], ["Alice", 33, True]))
    assert Record(zip(["name", "age"], ["Alice", 33])) == r[0:2]


def test_record_get_by_index():
    r = Record(zip(["name", "age", "married"], ["Alice", 33, True]))
    assert r[0] == "Alice"


def test_record_get_by_name():
    r = Record(zip(["name", "age", "married"], ["Alice", 33, True]))
    assert r["name"] == "Alice"


def test_record_get_by_out_of_bounds_index():
    r = Record(zip(["name", "age", "married"], ["Alice", 33, True]))
    assert r[9] is None



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import timedelta
from unittest import TestCase

from neo4j.time import Duration


class DurationTestCase(TestCase):

    def test_zero(self):
        d = Duration()
        self.assertEqual(d.months, 0)
        self.assertEqual(d.days, 0)
        self.assertEqual(d.seconds, 0)
        self.assertEqual(d.subseconds, 0.0)
        self.assertEqual(d.years_months_days, (0, 0, 0))
        self.assertEqual(d.hours_minutes_seconds, (0, 0, 0.0))
        self.assertFalse(bool(d))

    def test_years_only(self):
        d = Duration(years=2)
        self.assertEqual(d.months, 24)
        self.assertEqual(d.days, 0)
        self.assertEqual(d.seconds, 0)
        self.assertEqual(d.subseconds, 0.0)
        self.assertEqual(d.years_months_days, (2, 0, 0))
        self.assertEqual(d.hours_minutes_seconds, (0, 0, 0.0))

    def test_months_only(self):
        d = Duration(months=20)
        self.assertEqual(d.months, 20)
        self.assertEqual(d.days, 0)
        self.assertEqual(d.seconds, 0)
        self.assertEqual(d.subseconds, 0.0)
        self.assertEqual(d.years_months_days, (1, 8, 0))
        self.assertEqual(d.hours_minutes_seconds, (0, 0, 0.0))

    def test_months_out_of_range(self):
        with self.assertRaises(ValueError):
            _ = Duration(months=(2**64))

    def test_weeks_only(self):
        d = Duration(weeks=4)
        self.assertEqual(d.months, 0)
        self.assertEqual(d.days, 28)
        self.assertEqual(d.seconds, 0)
        self.assertEqual(d.subseconds, 0.0)
        self.assertEqual(d.years_months_days, (0, 0, 28))
        self.assertEqual(d.hours_minutes_seconds, (0, 0, 0.0))

    def test_days_only(self):
        d = Duration(days=40)
        self.assertEqual(d.months, 0)
        self.assertEqual(d.days, 40)
        self.assertEqual(d.seconds, 0)
        self.assertEqual(d.subseconds, 0.0)
        self.assertEqual(d.years_months_days, (0, 0, 40))
        self.assertEqual(d.hours_minutes_seconds, (0, 0, 0.0))

    def test_days_out_of_range(self):
        with self.assertRaises(ValueError):
            _ = Duration(days=(2**64))

    def test_hours_only(self):
        d = Duration(hours=10)
        self.assertEqual(d.months, 0)
        self.assertEqual(d.days, 0)
        self.assertEqual(d.seconds, 36000)
        self.assertEqual(d.subseconds, 0.0)
        self.assertEqual(d.years_months_days, (0, 0, 0))
        self.assertEqual(d.hours_minutes_seconds, (10, 0, 0.0))

    def test_minutes_only(self):
        d = Duration(minutes=90.5)
        self.assertEqual(d.months, 0)
        self.assertEqual(d.days, 0)
        self.assertEqual(d.seconds, 5430)
        self.assertEqual(d.subseconds, 0.0)
        self.assertEqual(d.years_months_days, (0, 0, 0))
        self.assertEqual(d.hours_minutes_seconds, (1, 30, 30.0))

    def test_seconds_only(self):
        d = Duration(seconds=123.456)
        self.assertEqual(d.months, 0)
        self.assertEqual(d.days, 0)
        self.assertEqual(d.seconds, 123)
        self.assertEqual(d.subseconds, 0.456)
        self.assertEqual(d.years_months_days, (0, 0, 0))
        self.assertEqual(d.hours_minutes_seconds, (0, 2, 3.456))

    def test_seconds_out_of_range(self):
        with self.assertRaises(ValueError):
            _ = Duration(seconds=(2**64))

    def test_subseconds_only(self):
        d = Duration(subseconds=123.456)
        self.assertEqual(d.months, 0)
        self.assertEqual(d.days, 0)
        self.assertEqual(d.seconds, 123)
        self.assertEqual(d.subseconds, 0.456)
        self.assertEqual(d.years_months_days, (0, 0, 0))
        self.assertEqual(d.hours_minutes_seconds, (0, 2, 3.456))

    def test_milliseconds_only(self):
        d = Duration(milliseconds=1234.567)
        self.assertEqual(d.months, 0)
        self.assertEqual(d.days, 0)
        self.assertEqual(d.seconds, 1)
        self.assertEqual(d.subseconds, 0.234567)
        self.assertEqual(d.years_months_days, (0, 0, 0))
        self.assertEqual(d.hours_minutes_seconds, (0, 0, 1.234567))

    def test_microseconds_only(self):
        d = Duration(microseconds=1234.567)
        self.assertEqual(d.months, 0)
        self.assertEqual(d.days, 0)
        self.assertEqual(d.seconds, 0)
        self.assertEqual(d.subseconds, 0.001234567)
        self.assertEqual(d.years_months_days, (0, 0, 0))
        self.assertEqual(d.hours_minutes_seconds, (0, 0, 0.001234567))

    def test_nanoseconds_only(self):
        d = Duration(nanoseconds=1234.567)
        self.assertEqual(d.months, 0)
        self.assertEqual(d.days, 0)
        self.assertEqual(d.seconds, 0)
        self.assertEqual(d.subseconds, 0.000001234)
        self.assertEqual(d.years_months_days, (0, 0, 0))
        self.assertEqual(d.hours_minutes_seconds, (0, 0, 0.000001234))

    def test_can_combine_years_months(self):
        t = Duration(years=5, months=3)
        self.assertEqual(t.months, 63)

    def test_can_combine_weeks_and_days(self):
        t = Duration(weeks=5, days=3)
        self.assertEqual(t.days, 38)

    def test_can_combine_hours_minutes_seconds(self):
        t = Duration(hours=5, minutes=4, seconds=3)
        self.assertEqual(t.seconds, 18243)

    def test_can_combine_seconds_and_subseconds(self):
        t = Duration(seconds=123.456, subseconds=0.321)
        self.assertEqual(t.seconds, 123)
        self.assertEqual(t.subseconds, 0.777)

    def test_full_positive(self):
        d = Duration(years=1, months=2, days=3, hours=4, minutes=5, seconds=6.789)
        self.assertEqual(d.months, 14)
        self.assertEqual(d.days, 3)
        self.assertEqual(d.seconds, 14706)
        self.assertEqual(d.subseconds, 0.789)
        self.assertEqual(d.years_months_days, (1, 2, 3))
        self.assertEqual(d.hours_minutes_seconds, (4, 5, 6.789))
        self.assertTrue(bool(d))

    def test_full_negative(self):
        d = Duration(years=-1, months=-2, days=-3, hours=-4, minutes=-5, seconds=-6.789)
        self.assertEqual(d.months, -14)
        self.assertEqual(d.days, -3)
        self.assertEqual(d.seconds, -14706)
        self.assertEqual(d.subseconds, -0.789)
        self.assertEqual(d.years_months_days, (-1, -2, -3))
        self.assertEqual(d.hours_minutes_seconds, (-4, -5, -6.789))
        self.assertTrue(bool(d))

    def test_negative_positive_negative(self):
        d = Duration(years=-1, months=-2, days=3, hours=-4, minutes=-5, seconds=-6.789)
        self.assertEqual(d.months, -14)
        self.assertEqual(d.days, 3)
        self.assertEqual(d.seconds, -14706)
        self.assertEqual(d.subseconds, -0.789)
        self.assertEqual(d.years_months_days, (-1, -2, 3))
        self.assertEqual(d.hours_minutes_seconds, (-4, -5, -6.789))

    def test_positive_negative_positive(self):
        d = Duration(years=1, months=2, days=-3, hours=4, minutes=5, seconds=6.789)
        self.assertEqual(d.months, 14)
        self.assertEqual(d.days, -3)
        self.assertEqual(d.seconds, 14706)
        self.assertEqual(d.subseconds, 0.789)
        self.assertEqual(d.years_months_days, (1, 2, -3))
        self.assertEqual(d.hours_minutes_seconds, (4, 5, 6.789))

    def test_add_duration(self):
        d1 = Duration(months=2, days=3, seconds=5.7)
        d2 = Duration(months=7, days=5, seconds=3.2)
        self.assertEqual(d1 + d2, Duration(months=9, days=8, seconds=8.9))

    def test_add_timedelta(self):
        d1 = Duration(months=2, days=3, seconds=5.7)
        td = timedelta(days=5, seconds=3.2)
        self.assertEqual(d1 + td, Duration(months=2, days=8, seconds=8.9))

    def test_add_object(self):
        with self.assertRaises(TypeError):
            _ = Duration(months=2, days=3, seconds=5.7) + object()

    def test_subtract_duration(self):
        d1 = Duration(months=2, days=3, seconds=5.7)
        d2 = Duration(months=7, days=5, seconds=3.2)
        self.assertEqual(d1 - d2, Duration(months=-5, days=-2, seconds=2.5))

    def test_subtract_timedelta(self):
        d1 = Duration(months=2, days=3, seconds=5.7)
        td = timedelta(days=5, seconds=3.2)
        self.assertEqual(d1 - td, Duration(months=2, days=-2, seconds=2.5))

    def test_subtract_object(self):
        with self.assertRaises(TypeError):
            _ = Duration(months=2, days=3, seconds=5.7) - object()

    def test_multiplication_by_int(self):
        d1 = Duration(months=2, days=3, seconds=5.7)
        i = 11
        self.assertEqual(d1 * i, Duration(months=22, days=33, seconds=62.7))

    def test_multiplication_by_float(self):
        d1 = Duration(months=2, days=3, seconds=5.7)
        f = 5.5
        self.assertEqual(d1 * f, Duration(months=11, days=16, seconds=31.35))

    def test_multiplication_by_object(self):
        with self.assertRaises(TypeError):
            _ = Duration(months=2, days=3, seconds=5.7) * object()

    def test_floor_division_by_int(self):
        d1 = Duration(months=11, days=33, seconds=55.77)
        i = 2
        self.assertEqual(d1 // i, Duration(months=5, days=16, seconds=27.0))

    def test_floor_division_by_object(self):
        with self.assertRaises(TypeError):
            _ = Duration(months=2, days=3, seconds=5.7) // object()

    def test_modulus_by_int(self):
        d1 = Duration(months=11, days=33, seconds=55.77)
        i = 2
        self.assertEqual(d1 % i, Duration(months=1, days=1, seconds=1.77))

    def test_modulus_by_object(self):
        with self.assertRaises(TypeError):
            _ = Duration(months=2, days=3, seconds=5.7) % object()

    def test_floor_division_and_modulus_by_int(self):
        d1 = Duration(months=11, days=33, seconds=55.77)
        i = 2
        self.assertEqual(divmod(d1, i), (Duration(months=5, days=16, seconds=27.0),
                                         Duration(months=1, days=1, seconds=1.77)))

    def test_floor_division_and_modulus_by_object(self):
        with self.assertRaises(TypeError):
            _ = divmod(Duration(months=2, days=3, seconds=5.7), object())

    def test_true_division_by_int(self):
        d1 = Duration(months=11, days=33, seconds=55.77)
        i = 2
        self.assertEqual(d1 / i, Duration(months=6, days=16, seconds=27.885))

    def test_true_division_by_float(self):
        d1 = Duration(months=11, days=33, seconds=55.77)
        f = 2.5
        self.assertEqual(d1 / f, Duration(months=4, days=13, seconds=22.308))

    def test_true_division_by_object(self):
        with self.assertRaises(TypeError):
            _ = Duration(months=2, days=3, seconds=5.7) / object()

    def test_unary_plus(self):
        d = Duration(months=11, days=33, seconds=55.77)
        self.assertEqual(+d, Duration(months=11, days=33, seconds=55.77))

    def test_unary_minus(self):
        d = Duration(months=11, days=33, seconds=55.77)
        self.assertEqual(-d, Duration(months=-11, days=-33, seconds=-55.77))

    def test_absolute(self):
        d = Duration(months=-11, days=-33, seconds=-55.77)
        self.assertEqual(abs(d), Duration(months=11, days=33, seconds=55.77))

    def test_str(self):
        self.assertEqual(str(Duration()), "PT0S")
        self.assertEqual(str(Duration(years=1, months=2)), "P1Y2M")
        self.assertEqual(str(Duration(years=-1, months=2)), "P-10M")
        self.assertEqual(str(Duration(months=-13)), "P-1Y-1M")
        self.assertEqual(str(Duration(months=2, days=3, seconds=5.7)), "P2M3DT5.7S")
        self.assertEqual(str(Duration(hours=12, minutes=34)), "PT12H34M")
        self.assertEqual(str(Duration(seconds=59)), "PT59S")
        self.assertEqual(str(Duration(seconds=0.123456789)), "PT0.123456789S")
        self.assertEqual(str(Duration(seconds=-0.123456789)), "PT-0.123456789S")

    def test_repr(self):
        d = Duration(months=2, days=3, seconds=5.7)
        self.assertEqual(repr(d), "Duration(months=2, days=3, seconds=5, subseconds=0.7)")

    def test_iso_format(self):
        self.assertEqual(Duration().iso_format(), "PT0S")
        self.assertEqual(Duration(years=1, months=2).iso_format(), "P1Y2M")
        self.assertEqual(Duration(years=-1, months=2).iso_format(), "P-10M")
        self.assertEqual(Duration(months=-13).iso_format(), "P-1Y-1M")
        self.assertEqual(Duration(months=2, days=3, seconds=5.7).iso_format(), "P2M3DT5.7S")
        self.assertEqual(Duration(hours=12, minutes=34).iso_format(), "PT12H34M")
        self.assertEqual(Duration(seconds=59).iso_format(), "PT59S")
        self.assertEqual(Duration(seconds=0.123456789).iso_format(), "PT0.123456789S")
        self.assertEqual(Duration(seconds=-0.123456789).iso_format(), "PT-0.123456789S")

    def test_from_iso_format(self):
        self.assertEqual(Duration(), Duration.from_iso_format("PT0S"))
        self.assertEqual(Duration(hours=12, minutes=34, seconds=56.789),
                         Duration.from_iso_format("PT12H34M56.789S"))
        self.assertEqual(Duration(years=1, months=2, days=3),
                         Duration.from_iso_format("P1Y2M3D"))
        self.assertEqual(Duration(years=1, months=2, days=3, hours=12, minutes=34, seconds=56.789),
                         Duration.from_iso_format("P1Y2M3DT12H34M56.789S"))



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import time
from unittest import TestCase

from pytz import (
    timezone,
    FixedOffset,
)

from neo4j.time import Time
from neo4j.time.arithmetic import (
    nano_add,
    nano_div,
)


timezone_us_eastern = timezone("US/Eastern")
timezone_utc = timezone("UTC")


class TimeTestCase(TestCase):

    def test_bad_attribute(self):
        t = Time(12, 34, 56.789)
        with self.assertRaises(AttributeError):
            _ = t.x

    def test_simple_time(self):
        t = Time(12, 34, 56.789)
        self.assertEqual(t.hour_minute_second, (12, 34, 56.789))
        self.assertEqual(t.ticks, 45296.789)
        self.assertEqual(t.hour, 12)
        self.assertEqual(t.minute, 34)
        self.assertEqual(t.second, 56.789)

    def test_midnight(self):
        t = Time(0, 0, 0)
        self.assertEqual(t.hour_minute_second, (0, 0, 0))
        self.assertEqual(t.ticks, 0)
        self.assertEqual(t.hour, 0)
        self.assertEqual(t.minute, 0)
        self.assertEqual(t.second, 0)

    def test_nanosecond_precision(self):
        t = Time(12, 34, 56.789123456)
        self.assertEqual(t.hour_minute_second, (12, 34, 56.789123456))
        self.assertEqual(t.ticks, 45296.789123456)
        self.assertEqual(t.hour, 12)
        self.assertEqual(t.minute, 34)
        self.assertEqual(t.second, 56.789123456)

    def test_str(self):
        t = Time(12, 34, 56.789123456)
        self.assertEqual(str(t), "12:34:56.789123456")

    def test_now_without_tz(self):
        t = Time.now()
        self.assertIsInstance(t, Time)

    def test_now_with_tz(self):
        t = Time.now(tz=timezone_us_eastern)
        self.assertIsInstance(t, Time)
        self.assertEqual(t.tzinfo, timezone_us_eastern)

    def test_utc_now(self):
        t = Time.utc_now()
        self.assertIsInstance(t, Time)

    def test_from_native(self):
        native = time(12, 34, 56, 789123)
        t = Time.from_native(native)
        self.assertEqual(t.hour, native.hour)
        self.assertEqual(t.minute, native.minute)
        self.assertEqual(t.second, nano_add(native.second, nano_div(native.microsecond, 1000000)))

    def test_to_native(self):
        t = Time(12, 34, 56.789123456)
        native = t.to_native()
        self.assertEqual(t.hour, native.hour)
        self.assertEqual(t.minute, native.minute)
        self.assertEqual(56.789123, nano_add(native.second, nano_div(native.microsecond, 1000000)))

    def test_iso_format(self):
        t = Time(12, 34, 56.789123456)
        self.assertEqual("12:34:56.789123456", t.iso_format())

    def test_iso_format_with_trailing_zeroes(self):
        t = Time(12, 34, 56.789)
        self.assertEqual("12:34:56.789000000", t.iso_format())

    def test_from_iso_format_hour_only(self):
        expected = Time(12, 0, 0)
        actual = Time.from_iso_format("12")
        self.assertEqual(expected, actual)

    def test_from_iso_format_hour_and_minute(self):
        expected = Time(12, 34, 0)
        actual = Time.from_iso_format("12:34")
        self.assertEqual(expected, actual)

    def test_from_iso_format_hour_minute_second(self):
        expected = Time(12, 34, 56)
        actual = Time.from_iso_format("12:34:56")
        self.assertEqual(expected, actual)

    def test_from_iso_format_hour_minute_second_milliseconds(self):
        expected = Time(12, 34, 56.123)
        actual = Time.from_iso_format("12:34:56.123")
        self.assertEqual(expected, actual)

    def test_from_iso_format_hour_minute_second_microseconds(self):
        expected = Time(12, 34, 56.123456)
        actual = Time.from_iso_format("12:34:56.123456")
        self.assertEqual(expected, actual)

    def test_from_iso_format_hour_minute_second_nanoseconds(self):
        expected = Time(12, 34, 56.123456789)
        actual = Time.from_iso_format("12:34:56.123456789")
        self.assertEqual(expected, actual)

    def test_from_iso_format_with_positive_tz(self):
        expected = Time(12, 34, 56.123456789, tzinfo=FixedOffset(754))
        actual = Time.from_iso_format("12:34:56.123456789+12:34")
        self.assertEqual(expected, actual)

    def test_from_iso_format_with_negative_tz(self):
        expected = Time(12, 34, 56.123456789, tzinfo=FixedOffset(-754))
        actual = Time.from_iso_format("12:34:56.123456789-12:34")
        self.assertEqual(expected, actual)

    def test_from_iso_format_with_positive_long_tz(self):
        expected = Time(12, 34, 56.123456789, tzinfo=FixedOffset(754))
        actual = Time.from_iso_format("12:34:56.123456789+12:34:56.123456")
        self.assertEqual(expected, actual)

    def test_from_iso_format_with_negative_long_tz(self):
        expected = Time(12, 34, 56.123456789, tzinfo=FixedOffset(-754))
        actual = Time.from_iso_format("12:34:56.123456789-12:34:56.123456")
        self.assertEqual(expected, actual)


def test_iso_format_with_time_zone_case_1():
    # python -m pytest tests/unit/time/test_time.py -s -v -k test_iso_format_with_time_zone_case_1
    expected = Time(7, 54, 2.129790999, tzinfo=timezone_utc)
    assert expected.iso_format() == "07:54:02.129790999+00:00"
    assert expected.tzinfo == FixedOffset(0)
    actual = Time.from_iso_format("07:54:02.129790999+00:00")
    assert expected == actual


def test_iso_format_with_time_zone_case_2():
    # python -m pytest tests/unit/time/test_time.py -s -v -k test_iso_format_with_time_zone_case_2
    expected = Time.from_iso_format("07:54:02.129790999+01:00")
    assert expected.tzinfo == FixedOffset(60)
    assert expected.iso_format() == "07:54:02.129790999+01:00"


def test_to_native_case_1():
    # python -m pytest tests/unit/time/test_time.py -s -v -k test_to_native_case_1
    t = Time(12, 34, 56.789123456)
    native = t.to_native()
    assert native.hour == t.hour
    assert native.minute == t.minute
    assert nano_add(native.second, nano_div(native.microsecond, 1000000)) == 56.789123
    assert native.tzinfo is None
    assert native.isoformat() == "12:34:56.789123"


def test_to_native_case_2():
    # python -m pytest tests/unit/time/test_time.py -s -v -k test_to_native_case_2
    t = Time(12, 34, 56.789123456, tzinfo=timezone_utc)
    native = t.to_native()
    assert native.hour == t.hour
    assert native.minute == t.minute
    assert nano_add(native.second, nano_div(native.microsecond, 1000000)) == 56.789123
    assert native.tzinfo == FixedOffset(0)
    assert native.isoformat() == "12:34:56.789123+00:00"


def test_from_native_case_1():
    # python -m pytest tests/unit/time/test_time.py -s -v -k test_from_native_case_1
    native = time(12, 34, 56, 789123)
    t = Time.from_native(native)
    assert t.hour == native.hour
    assert t.minute == native.minute
    assert t.second == nano_add(native.second, nano_div(native.microsecond, 1000000))
    assert t.tzinfo is None


def test_from_native_case_2():
    # python -m pytest tests/unit/time/test_time.py -s -v -k test_from_native_case_2
    native = time(12, 34, 56, 789123, FixedOffset(0))
    t = Time.from_native(native)
    assert t.hour == native.hour
    assert t.minute == native.minute
    assert t.second == nano_add(native.second, nano_div(native.microsecond, 1000000))
    assert t.tzinfo == FixedOffset(0)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import struct
from unittest import TestCase

from neo4j.data import DataDehydrator
from neo4j.packstream import Packer
from neo4j.spatial import WGS84Point


class WGS84PointTestCase(TestCase):

    def test_alias(self):
        x, y, z = 3.2, 4.0, -1.2
        p = WGS84Point((x, y, z))
        self.assert_(hasattr(p, "longitude"))
        self.assertEqual(p.longitude, x)
        self.assert_(hasattr(p, "latitude"))
        self.assertEqual(p.latitude, y)
        self.assert_(hasattr(p, "height"))
        self.assertEqual(p.height, z)

    def test_dehydration_3d(self):
        coordinates = (1, -2, 3.1)
        p = WGS84Point(coordinates)

        dehydrator = DataDehydrator()
        buffer = io.BytesIO()
        packer = Packer(buffer)
        packer.pack(dehydrator.dehydrate((p,))[0])
        self.assertEqual(
            buffer.getvalue(),
            b"\xB4Y" +
            b"\xC9" + struct.pack(">h", 4979) +
            b"".join(map(lambda c: b"\xC1" + struct.pack(">d", c), coordinates))
        )

    def test_dehydration_2d(self):
        coordinates = (.1, 0)
        p = WGS84Point(coordinates)

        dehydrator = DataDehydrator()
        buffer = io.BytesIO()
        packer = Packer(buffer)
        packer.pack(dehydrator.dehydrate((p,))[0])
        self.assertEqual(
            buffer.getvalue(),
            b"\xB3X" +
            b"\xC9" + struct.pack(">h", 4326) +
            b"".join(map(lambda c: b"\xC1" + struct.pack(">d", c), coordinates))
        )



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) "Neo4j"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
from uuid import uuid4

from neo4j.work.simple import unit_of_work
from neo4j.exceptions import (
    Neo4jError,
    ClientError,
)

# python -m pytest tests/integration/test_tx_functions.py -s -v


def test_simple_read(session):

    def work(tx):
        return tx.run("RETURN 1").single().value()

    value = session.read_transaction(work)
    assert value == 1


def test_read_with_arg(session):

    def work(tx, x):
        return tx.run("RETURN $x", x=x).single().value()

    value = session.read_transaction(work, x=1)
    assert value == 1


def test_read_with_arg_and_metadata(session):

    @unit_of_work(timeout=25, metadata={"foo": "bar"})
    def work(tx):
        return tx.run("CALL dbms.getTXMetaData").single().value()

    try:
        value = session.read_transaction(work)
    except ClientError:
        pytest.skip("Transaction metadata and timeout only supported in Neo4j EE 3.5+")
    else:
        assert value == {"foo": "bar"}


def test_simple_write(session):

    def work(tx):
        return tx.run("CREATE (a {x: 1}) RETURN a.x").single().value()

    value = session.write_transaction(work)
    assert value == 1


def test_write_with_arg(session):

    def work(tx, x):
        return tx.run("CREATE (a {x: $x}) RETURN a.x", x=x).single().value()

    value = session.write_transaction(work, x=1)
    assert value == 1


def test_write_with_arg_and_metadata(session):

    @unit_of_work(timeout=25, metadata={"foo": "bar"})
    def work(tx, x):
        return tx.run("CREATE (a {x: $x}) RETURN a.x", x=x).single().value()

    try:
        value = session.write_transaction(work, x=1)
    except ClientError:
        pytest.skip("Transaction metadata and timeout only supported in Neo4j EE 3.5+")
    else:
        assert value == 1


def test_error_on_write_transaction(session):

    def f(tx, uuid):
        tx.run("CREATE (a:Thing {uuid:$uuid})", uuid=uuid), uuid4()

    with pytest.raises(TypeError):
        session.write_transaction(f)


def test_retry_logic(driver):
    # python -m pytest tests/integration/test_tx_functions.py -s -v -k test_retry_logic

    pytest.global_counter = 0

    def get_one(tx):
        result = tx.run("UNWIND [1,2,3,4] AS x RETURN x")
        records = list(result)
        pytest.global_counter += 1

        if pytest.global_counter < 3:
            database_unavailable = Neo4jError.hydrate(message="The database is not currently available to serve your request, refer to the database logs for more details. Retrying your request at a later time may succeed.", code="Neo.TransientError.Database.DatabaseUnavailable")
            raise database_unavailable

        return records

    with driver.session() as session:
        records = session.read_transaction(get_one)

        assert pytest.global_counter == 3

    del pytest.global_counter
